# pysam versioning information
__version__ = "0.11.2.2"

# TODO: upgrade number
__samtools_version__ = "1.4.1"

# TODO: upgrade code and number
__bcftools_version__ = "1.4.1"

__htslib_version__ = "1.4.1"
